#include "managers.h"
#include "global.h"
#include "xpms.h"
#include <algorithm>
#include <stdexcept>


/*
 *  Wave
 */
 
Wave::Wave(const WaveData & data)
: pix_         (data.alien_name)
, wait_        (data.wait)
, wave_size_pc_(data.wave_size_pc)
, arrival_     (data.arrival, data.arrival + data.n_rcoords)
, mirrored_    (data.mirrored)
{}


int Wave::Size() const
{
	return X11::Inst().WindowWidth() * wave_size_pc_ /
	       g_aliens_hspacing          / 100;
}



/*
 *  AliensManager
 */

int AliensManager::bonus_wait_;


AliensManager::AliensManager(int level_number, const WaveData * wavesdata,
                             int nwaves, int max_level)
: level_number_     (level_number)
, wave_number_      (0)
, max_wave_size_    (0)
, speed_            (level_number / (max_level + 1) + 3)
, base_cruise_      (Coord(0, 10))
, base_cruise_speed_(1)
, next_creation_    (new_wave)
{
	if (level_number < 1 || !wavesdata || nwaves < 1) {
		throw std::runtime_error("AliensManager data error");
	}
	for (int i (0); i < nwaves; ++i) {
		waves_.push_back(new Wave(wavesdata[i]));
		if (waves_[i]->Size() > max_wave_size_) {
			max_wave_size_ = waves_[i]->Size();
		}
	}
	next_creation_wait_ = waves_[0]->Wait();
	if (level_number == 1) bonus_wait_ = 20;
}


template <class Ctr> inline void Delete(typename Ctr::value_type ptr)
	{ delete ptr; }

AliensManager::~AliensManager()
{
	std::for_each(waves_.begin(), waves_.end(), Delete<WavesCtr>);
	std::for_each(aliens_.begin(), aliens_.end(), Delete<AliensCtr>);
}


void AliensManager::Move() {
	base_cruise_.x += base_cruise_speed_;
	if (base_cruise_speed_ < 0 && base_cruise_.x <= g_alien_width / 2 ||
	    base_cruise_speed_ > 0 && base_cruise_.x >=
	    max_wave_size_ * g_aliens_hspacing + g_alien_width / 2 +
	    (g_aliens_hspacing - g_alien_width)) {
		base_cruise_speed_ *= -1;
	}

	for (std::list<Alien*>::iterator it (aliens_.begin());
	     it != aliens_.end(); ++it) {
		(*it)->Move();
	}

	if (next_creation_wait_ == 0) Creation();
	else --next_creation_wait_;
}


std::list<Bullet*> AliensManager::Fire(int nfired) const
{
	std::list<Bullet*> res;
	for (AliensCtr::const_iterator it (aliens_.begin());
	     it != aliens_.end() && nfired < 3 * level_number_ * max_wave_size_ / 10;
	     ++it) {
		if ((*it)->Stage() != Trajectory::cruising &&
		    (*it)->Stage() != Trajectory::joining) {
			if (rand() % 50 < speed_) {
				res.push_back((*it)->Fire());
				++nfired;
			}
		}
	}
	return res;
}


std::pair<Explosion*, Bonus*> AliensManager::Collision(const Bullet * bullet)
{
	std::pair<Explosion*, Bonus*> res (0, 0);
	for (AliensCtr::iterator a (aliens_.begin());
	     a != aliens_.end(); ++a) {
	     	const Coord alien_pos ((*a)->Position());
		if (bullet->Collide(alien_pos,
		                    Coord((*a)->Width(), (*a)->Height()))) {
			XColor color;
			color.red = color.green = 65535; color.blue = 54*1024;
			res.first = new Explosion((*a)->Position(),
			                          Coord(0,-4), color);
			(*a)->Erase();
			delete *a;
			aliens_.erase(a);
			res.second = CreateBonusMaybe(alien_pos);
			break;
		}
	}
	return res;
}


void AliensManager::Creation()
{
	switch (next_creation_) {
	case new_wave:
		wave_alien_number_ = 0;
		next_creation_ = new_alien;
		// continue with new_alien
	case new_alien: {
		const Wave *const wave (waves_[wave_number_]);
		const Pix *const pix (wave->AlienPix());
		const Coord delta_cruise
			((wave_alien_number_ +
			  (max_wave_size_ - wave->Size()) / 2) *
			 g_aliens_hspacing,
			 (waves_.size() - wave_number_) * g_aliens_vspacing);
		const bool mirrored
			(wave->Mirrored() && wave_alien_number_ % 2 == 0);
		const Trajectory trajectory
			(wave->Arrival(), mirrored, base_cruise_, delta_cruise);
		Alien *const alien (new Alien(pix, trajectory, speed_));
		aliens_.push_back(alien);

		if (++wave_alien_number_ < wave->Size()) {
			next_creation_wait_ = 30 / speed_;
		} else {
			if (++wave_number_ < waves_.size()) {
				next_creation_ = new_wave;
				next_creation_wait_ = waves_[wave_number_]->Wait();
			} else {
				next_creation_ = new_attack;
				next_creation_wait_ =
					std::max(0, 750 - level_number_ * 50);
			}
		}
		break;
	}
	case new_attack: {
		int nb_attacking (0);
		for (AliensCtr::iterator it (aliens_.begin());
		     it != aliens_.end(); ++it) {
			if ((*it)->Stage() == Trajectory::attacking) ++nb_attacking;
		}
		AliensCtr::iterator a (aliens_.begin());
		while (nb_attacking < level_number_ * 3 && a != aliens_.end()) {
			if ((*a)->Stage() == Trajectory::cruising) {
				(*a)->BuildAttack();
				++nb_attacking;
			}
			++a;
		}
		next_creation_wait_ = 5;
		break;
	}
	} // switch
}


Bonus * AliensManager::CreateBonusMaybe(Coord pos)
{
	if (--bonus_wait_ > 0) return 0;
	bonus_wait_ = max_wave_size_ * 6;
	switch (rand() % 5) {
	case 0:
		return new Bonus(PixKeeper::Instance().Get(bonus_speed_xpm),
		                 pos, Bonus::extra_speed);
	case 1: case 2: case 3:
		return new Bonus(PixKeeper::Instance().Get(bonus_fire_xpm),
		                 pos, Bonus::extra_fire);
	default:
		return new Bonus(PixKeeper::Instance().Get(bonus_shield_xpm),
		                 pos, Bonus::extra_shield);
	}
}



/*
 *  BulletsManager
 */
 
BulletsManager::~BulletsManager()
{
	for (BulletsCtr::iterator it (bullets_.begin());
	     it != bullets_.end(); ++it) {
		delete *it;
	}	
}


void BulletsManager::Move()
{
	BulletsCtr::iterator it (bullets_.begin());
	while (it != bullets_.end()) {
		(*it)->Move();
		if ((*it)->Out()) it = Destroy(it);
		else ++it;
	}	
}


BulletsManager::BulletsCtr::iterator BulletsManager::Destroy(BulletsCtr::iterator it)
{
	delete *it;
	return bullets_.erase(it);
}


/*
 *  BonusManager
 */
 
BonusManager::~BonusManager()
{
	for (BonusCtr::iterator it (bonuses_.begin());
	     it != bonuses_.end(); ++it) {
		delete *it;
	}
}


void BonusManager::Move()
{
	BonusCtr::iterator it (bonuses_.begin());
	while (it != bonuses_.end()) {
		(*it)->Move();
		if ((*it)->Out()) it = Destroy(it);
		else ++it;
	}
}


BonusManager::BonusCtr::iterator BonusManager::Destroy(BonusCtr::iterator it)
{
	delete *it;
	return bonuses_.erase(it);
}



/*
 *  ExplosionsManager
 */

ExplosionsManager::~ExplosionsManager()
{
	for (ExplosionsCtr::iterator it (explosions_.begin());
	     it != explosions_.end(); ++it) {
		delete *it;
	}
}


void ExplosionsManager::Move()
{
	ExplosionsCtr::iterator it (explosions_.begin());
	while (it != explosions_.end()) {
		(*it)->Move();
		if ((*it)->Finished()) it = Destroy(it);
		else ++it;
	}
}


ExplosionsManager::ExplosionsCtr::iterator
ExplosionsManager::Destroy(ExplosionsCtr::iterator it)
{
	delete *it;
	return explosions_.erase(it);
}

