#ifndef SPRITES_H
#define SPRITES_H

#include "gfxinterface.h"
#include "pix.h"
#include "xpms.h"
#include <list>
#include <vector>


// One trajectory is given to each Alien. The trajectory computes the next
// position of the Alien. The Trajectory can also build an attack on demand.
// A trajectory has 4 stages:
// - an arriving stage consisting of passing through given RCoord's;
// - a joining stage to join the cruising position;
// - a cruising stage at position given by a changing base and a fixed delta;
// - an attacking stage consisting of passing through random RCoord's.
// The trajectory records the position of an alien.
class Trajectory {
public:
	enum stage_t { arriving, joining, cruising, attacking };
private:
	stage_t stage_;
	const std::vector<RCoord> & arrival_;
	std::vector<RCoord>::const_iterator arrival_it_;
	const bool mirrored_;
	const Coord & base_cruise_;
	const Coord delta_cruise_;
	std::vector<RCoord> attack_;
	std::vector<RCoord>::const_iterator attack_it_;
	Coord pos_;
public:
	Trajectory(const std::vector<RCoord> & arrival,
	           bool mirrored,
	           const Coord & base_cruise,
	           Coord delta_cruise);
	stage_t Stage() const { return stage_; }
	Coord Position() const { return pos_; }
	void NextPosition(int speed);
	void BuildAttack();
};


// An explosion is made of color changing dots.
class Explosion {
	XColor color_;
	int moves_;
	const int duration_;
	typedef std::vector<Coord> ParticlesCtr;
	ParticlesCtr particles_, particles_speed_;
public:
	Explosion(Coord pos, Coord speed, XColor color, int duration = 24);
	void Move();
	bool Finished() const { return moves_ >= duration_; }
private:
	void DrawAll(X11::Color color) const;
};


// Projectile is the base class for constant speed moving objects.
class Projectile {
	const Pix * pix_;
	Coord pos_;
	const Coord speed_;
public:
	Projectile(const Pix * pix, Coord pos, Coord speed);
	virtual ~Projectile() { pix_->Erase(pos_); }
	void Move();
	bool Collide(Coord pos2, Coord size) const;
	bool Out() const { return pos_.YOut(); }
};


// A bullet is displayed by a pix and moves straight at a constant 2D speed.
class Bullet : public Projectile {
public:
	Bullet(const Pix * pix, Coord pos, Coord speed)
		: Projectile(pix, pos, speed) {}
};


// A bonus is displayed by a pix and drops at a constant speed.
class Bonus : public Projectile {
public:
	enum bonus_t { extra_speed, extra_fire, extra_shield };
private:
	const bonus_t type_;
public:
	Bonus(const Pix * pix, Coord pos, bonus_t type, Coord speed = Coord(0,1))
		: Projectile(pix, pos, speed), type_(type) {}
	bonus_t Type() const { return type_; }
};


// An alien vessel is displayed by a pix at a position given by its trajectory.
// It can fire bullets (dropping bombs). It can collide with player bullets.
class Alien {
	const Pix *const pix_;
	const Pix *const bomb_pix_;
	Trajectory trajectory_;
	const int speed_;
public:
	Alien(const Pix * pix, const Trajectory & trajectory, int speed);
	void Move();
	void Erase() const { pix_->Erase(Position()); }
	Bullet * Fire() const;
	Coord Position() const { return trajectory_.Position(); }
	int Width()  const { return pix_->Width();  }
	int Height() const { return pix_->Height(); }
	Trajectory::stage_t Stage() const { return trajectory_.Stage(); }
	void BuildAttack() { trajectory_.BuildAttack(); }
};


// The player ship is displayed by a pix and moves on the bottom of the window.
// It can fire bullets. It can collide with alien bullets (bombs) and bonuses.
// The shield is decremented by each collision.
class Player {
	Player(const Player &); // no copy
	Player & operator=(const Player &); // no copy

	const Pix *const pix_;
	const Pix *const bullet_pix_;
	const int below_y_;
	Coord pos_;
	int shield_;
	int speed_;
	int fire_wait_, fire_interval_;
	int multi_fire_;
public:
	Player(int below_y);
	void Move(int direction);
	int Width()  const { return pix_->Width(); }
	int Height() const { return pix_->Height(); }
	std::list<Bullet*> Fire();
	Explosion * Collision(const Bullet *);
	bool Collision(const Bonus *);
	int Shield() const { return shield_; }
	void ExtraFire();
	void ExtraShield() { ++shield_; }
	void ExtraSpeed() { ++speed_; }
private:
	void Fire(std::list<Bullet*> * res) const;
	int PosY() const;
};


#endif
